﻿//
//  コンピュータアニメーション特論
//  複数オブジェクトの位置・向きをマウスで操作するためのクラス
//


#include "ObjectLayout.h"

// GLUTヘッダファイルのインクルード
#include <GL/glut.h>

// vecmathヘッダファイルのインクルード
#include "vecmath_gl.h"

// 標準算術関数・定数の定義
#define  _USE_MATH_DEFINES
#include <math.h>



//
//  コンストラクタ
//
ObjectLayout::ObjectLayout()
{
	curr_object = -1;
	curr_operation = OP_W_TRANSLATION;
	active_handle = H_NONE;
	on_control = false;

	// 描画オプションの設定
	render_option.box_color.set( 0.8f, 0.8f, 0.0f );
	render_option.selected_box_color.set( 0.8f, 0.0f, 0.8f );

	render_option.box_width = 1.0f;
	render_option.axis_width = 2.0f;
	render_option.selected_axis_width = 6.0f;

	render_option.x_axis_color.set( 1.0f, 0.0f, 0.0f );
	render_option.y_axis_color.set( 0.0f, 1.0f, 0.0f );
	render_option.z_axis_color.set( 0.0f, 0.0f, 1.0f );

	render_option.change_axis_length = false;
	render_option.enable_xray_mode = true;
	render_option.enable_smooth = true;

	// デフォルトのオブジェクト情報を設定
	default_object.pos.set( 0.0f, 0.0f, 0.0f );
	default_object.ori.setIdentity();
	default_object.size.set( 1.0f, 1.0f, 1.0f );
	default_object.axis_length = 0.8f; 
}



//
//  オブジェクトの位置・向きを表す変換行列を返す
//
Matrix4f &  ObjectLayout::GetFrame( int no )
{
	Object &  o = objects[ no ];
	o.frame.setIdentity();
	o.frame.set( o.ori, o.pos, 1.0f );
	return  o.frame;
}


//
//  オブジェクトの位置・向きを表す変換行列を返す
//
Matrix4f &  ObjectLayout::GetTransposedFrame( int no )
{
	Object &  o = objects[ no ];
	o.frame.setIdentity();
	o.frame.set( o.ori, o.pos, 1.0f );
	o.frame.transpose();
	return  o.frame;
}


//
//  現在の操作モードを表す文字列を返す
//
const char *  ObjectLayout::GetOperationMode()
{
	// 各操作モードを表す文字列
	static const char *  operation_name[] = {
		"Translation in World Coordinates", "Translation in Model Coordinates", 
		"Rotation in World Coordinates", "Rotation in Model Coordinates" };

	// オブジェクトが選択されていれば操作モードを表す文字列を返す
	if ( curr_object != -1 )
		return  operation_name[ curr_operation ];

	return  NULL;
}



//////////////////////////////////////////////////////////////////////////////
//
//  オブジェクトの操作
//


//
//  オブジェクトの追加
//
int  ObjectLayout::AddObject()
{
	objects.push_back( default_object );

	// スクリーン座標系に投影した座標を計算
	ProjectObjects();

	return  objects.size() - 1;
}


//
//  オブジェクトの削除
//
int  ObjectLayout::DeleteObject()
{
	objects.pop_back();
	if ( curr_object >= objects.size() )
		curr_object = -1;
	return  objects.size();
}

int  ObjectLayout::DeleteAllObjects()
{
	objects.clear();
	curr_object = -1;
	return  0;
}



//
//  オブジェクトの情報設定
//
void  ObjectLayout::SetObjectPos( int no, const Point3f & p )
{
	objects[ no ].pos = p;
}

void  ObjectLayout::SetObjectOri( int no, const Matrix3f & o )
{
	objects[ no ].ori = o;
}

void  ObjectLayout::SetObjectFrame( int no, const Matrix4f & f )
{
	Vector3f  v;
	f.get( &v );
	objects[ no ].pos = v;

	f.get( &objects[ no ].ori );
}

void  ObjectLayout::SetObjectSize( int no, const Vector3f & s )
{
	objects[ no ].size = s;
	objects[ no ].axis_length = ( (s.x > s.y) ? ((s.x > s.z) ? s.x : s.z) : ((s.y > s.z) ? s.y : s.z) ) * 0.8f; 
}

void  ObjectLayout::SetObjectSize( int no, float s )
{
	objects[ no ].size.set( s, s, s );
	objects[ no ].axis_length = s * 0.8f; 
}



//////////////////////////////////////////////////////////////////////////////
//
//  イベントハンドラ
//


//
//  オブジェクト位置・視点の変更を通知
//
void  ObjectLayout::Update()
{
	// 全オブジェクトをスクリーン座標系に投影
	ProjectObjects();
}


//
//  マウスのボタンが押された時の処理
//
void  ObjectLayout::OnMouseDown( int mx, int my )
{
	// オブジェクトが選択されており、オブジェクトの上にマウスがあれば、操作モードを変更
	if ( ( curr_object != -1 ) && ( active_handle == H_MODE_CHANGE ) )
	{
		if ( curr_operation == OP_W_TRANSLATION )
			curr_operation = OP_M_ROTATION;
		else if ( curr_operation == OP_M_ROTATION )
			curr_operation = OP_M_TRANSLATION;
		else if ( curr_operation == OP_M_TRANSLATION )
			curr_operation = OP_W_ROTATION;
		else
			curr_operation = OP_W_TRANSLATION;
	}
	// オブジェクトが選択されており、ハンドルの上にマウスがあれば、操作を開始
	else if ( ( curr_object != -1 ) && ( active_handle != H_NONE ) )
	{
		on_control = true;

		// 操作に対応するマウス座標軸を判定
		// ワールド座標軸orモデル座標軸での移動
		if ( ( curr_operation == OP_W_TRANSLATION ) || ( curr_operation == OP_M_TRANSLATION ) )
		{
			// 移動を行う軸がスクリーン上で横方向と縦方向のどちらに長いかにより判定
			Point2f  handle_delta;
			handle_delta = handle_pos[ active_handle ] - handle_pos[ 0 ];
			if ( fabs( handle_delta.x ) > fabs( handle_delta.y ) )
			{
				// 軸の方向により符号を判定
				if ( handle_delta.x > 0.0f )
					mouse_control = MOUSE_PLUS_X;
				else
					mouse_control = MOUSE_MINUS_X;
			}
			else
			{
				// 軸の方向により符号を判定
				if ( handle_delta.y > 0.0f )
					mouse_control = MOUSE_PLUS_Y;
				else
					mouse_control = MOUSE_MINUS_Y;
			}
		}
		// ワールド座標軸orモデル座標軸での回転
		else if ( ( curr_operation == OP_W_ROTATION ) || ( curr_operation == OP_M_ROTATION ) )
		{
			// 回転を行う方向を取得
			Point2f  handle_delta;
			switch ( active_handle )
			{
			  case  H_M_XY_PLANE:
				handle_delta.sub( handle_pos[ H_M_Y_AXIS ], handle_pos[ H_M_X_AXIS ] );  break;
			  case  H_M_YZ_PLANE:
				handle_delta.sub( handle_pos[ H_M_Z_AXIS ], handle_pos[ H_M_Y_AXIS ] );  break;
			  case  H_M_ZX_PLANE:
				handle_delta.sub( handle_pos[ H_M_X_AXIS ], handle_pos[ H_M_Z_AXIS ] );  break;
			  case  H_W_XY_PLANE:
				handle_delta.sub( handle_pos[ H_W_Y_AXIS ], handle_pos[ H_W_X_AXIS ] );  break;
			  case  H_W_YZ_PLANE:
				handle_delta.sub( handle_pos[ H_W_Z_AXIS ], handle_pos[ H_W_Y_AXIS ] );  break;
			  case  H_W_ZX_PLANE:
				handle_delta.sub( handle_pos[ H_W_X_AXIS ], handle_pos[ H_W_Z_AXIS ] );  break;
			}

			// 回転を行う方向がスクリーン上で横方向と縦方向のどちらに長いかにより判定
			if ( fabs( handle_delta.x ) > fabs( handle_delta.y ) )
			{
				// 回転を行う方向の向きにより符号を判定
				if ( handle_delta.x > 0.0f )
					mouse_control = MOUSE_PLUS_X;
				else
					mouse_control = MOUSE_MINUS_X;
			}
			else
			{
				// 回転を行う方向の向きにより符号を判定
				if ( handle_delta.y > 0.0f )
					mouse_control = MOUSE_PLUS_Y;
				else
					mouse_control = MOUSE_MINUS_Y;
			}
		}

		// 現在のマウス座標を記録
		last_mouse_x = mx;
		last_mouse_y = my;
	}
	// オブジェクトを選択
	else
	{
		// クリック位置に近いオブジェクトを探索
		int  find_object = -1;
		find_object = FindObject( mx, my, 10 );
		curr_object = find_object;

		// ハンドラのスクリーン座標を計算
		ProjectObjects();
	}
}


//
//  マウスのボタンが離された時の処理
//
void  ObjectLayout::OnMouseUp( int mx, int my )
{
	// 現在の操作を解除する
	on_control = false;
}


//
//  マウス移動時の処理
//
void  ObjectLayout::OnMoveMouse( int mx, int my )
{
	const float  pos_scale = 0.01f;
	const float  angle_scale = 0.01f;

	// オブジェクトの位置・向きをコントロール
	if ( on_control )
	{
		Object &  o = objects[ curr_object ];
		int  dx, dy;
		dx = mx - last_mouse_x;
		dy = my - last_mouse_y;

		// マウス座標変化に応じて移動量を計算
		if ( ( curr_operation == OP_W_TRANSLATION ) || ( curr_operation == OP_M_TRANSLATION ) )
		{
			// 移動を行う軸を取得
			Vector3f  vec;
			switch ( active_handle )
			{
			  case  H_W_X_AXIS:
				vec.set( 1.0f, 0.0f, 0.0f );  break;
			  case  H_W_Y_AXIS:
				vec.set( 0.0f, 1.0f, 0.0f );  break;
			  case  H_W_Z_AXIS:
				vec.set( 0.0f, 0.0f, 1.0f );  break;
			  case  H_M_X_AXIS:
				o.ori.getColumn( 0, &vec );  break;
			  case  H_M_Y_AXIS:
				o.ori.getColumn( 1, &vec );  break;
			  case  H_M_Z_AXIS:
				o.ori.getColumn( 2, &vec );  break;
			}

			// 操作に対応するマウス座標軸に応じて移動を適用
			if ( mouse_control == MOUSE_PLUS_X )
				o.pos.scaleAdd( dx * pos_scale, vec, o.pos );
			else if ( mouse_control == MOUSE_MINUS_X )
				o.pos.scaleAdd( -dx * pos_scale, vec, o.pos );
			else if ( mouse_control == MOUSE_PLUS_Y )
				o.pos.scaleAdd( dy * pos_scale, vec, o.pos );
			else if ( mouse_control == MOUSE_MINUS_Y )
				o.pos.scaleAdd( -dy * pos_scale, vec, o.pos );
		}

		// マウス座標変化に応じて回転量を計算
		if ( ( curr_operation == OP_W_ROTATION ) || ( curr_operation == OP_M_ROTATION ) )
		{
			// 操作に対応するマウス座標軸に応じて回転角度を計算
			float  angle = 0.0f;
			if ( mouse_control == MOUSE_PLUS_X )
				angle = dx * angle_scale;
			else if ( mouse_control == MOUSE_PLUS_Y )
				angle = dy * angle_scale;
			else if ( mouse_control == MOUSE_MINUS_X )
				angle = -dx * angle_scale;
			else if ( mouse_control == MOUSE_MINUS_Y )
				angle = -dy * angle_scale;

			// 回転行列を適用
			Matrix3f  rot;
			switch ( active_handle )
			{
			  case  H_M_XY_PLANE:
				rot.rotZ( angle );  o.ori.mul( o.ori, rot );  break;
			  case  H_M_YZ_PLANE:
				rot.rotX( angle );  o.ori.mul( o.ori, rot );  break;
			  case  H_M_ZX_PLANE:
				rot.rotY( angle );  o.ori.mul( o.ori, rot );  break;
			  case  H_W_XY_PLANE:
				rot.rotZ( angle );  o.ori.mul( rot, o.ori );  break;
			  case  H_W_YZ_PLANE:
				rot.rotX( angle );  o.ori.mul( rot, o.ori );  break;
			  case  H_W_ZX_PLANE:
				rot.rotY( angle );  o.ori.mul( rot, o.ori );  break;
			}
		}

		// オブジェクト・ハンドルのスクリーン上の座標を更新
		ProjectObjects();

		// 現在のマウス座標を記録
		last_mouse_x = mx;
		last_mouse_y = my;
	}

	// マウス座標に近いハンドルを選択
	else if ( curr_object != -1 )
	{
		// マウス座標に近いハンドルを検索
		HandleEnum  handle = FindHandle( mx, my, curr_operation, 10 );

		// 移動操作モードはハンドルを
		active_handle = handle;
	}

	// マウス座標に近いオブジェクトを記録
	else
	{
	}
}


//
//  オブジェクトの操作情報の描画
//
void  ObjectLayout::Render()
{
	// 現在のレンダリング設定を記録
	GLboolean  b_lighting, b_depth_test, b_blend, b_line_smooth;
	glGetBooleanv( GL_LIGHTING, &b_lighting );
	glGetBooleanv( GL_DEPTH_TEST, &b_depth_test );
	glGetBooleanv( GL_BLEND, &b_blend );
	glGetBooleanv( GL_LINE_SMOOTH, &b_line_smooth );

	// 描画オプションに応じてレンダリング設定
	const RenderOption &  ro = render_option;
	glDisable( GL_LIGHTING );
	if ( ro.enable_smooth )
	{
		glEnable( GL_BLEND );
		glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
		glEnable( GL_LINE_SMOOTH );
	}

	// 選択オブジェクトを描画
	if ( curr_object != -1 )
	{
		Object &  o = objects[ curr_object ];
		Matrix4f  frame;
		frame.set( o.ori, o.pos, 1.0f );

		// 各軸の長さ（全軸で共通の長さを使用 or 物体のサイズを使用）
		float  sx, sy, sz;
		if ( ro.change_axis_length )
		{
			sx = o.size.x;  sy = o.size.y;  sz = o.size.z;
		}
		else
		{
			sx = o.axis_length;  sy = o.axis_length;  sz = o.axis_length;
		}

		// バウンディングボックスはＺバッファを考慮して描画
		glEnable( GL_DEPTH_TEST );

		// オブジェクトのバウンディングボックスを描画
		glPushMatrix();
			glMultMatrixf( frame );
			glScalef( o.size.x, o.size.y, o.size.z );
			glColor3f( ro.selected_box_color );
			glLineWidth( 1.0f );
			glutWireCube( 1.0f );
		glPopMatrix();

		// 座標系の描画でＺバッファを考慮するかどうかを設定
		if ( ro.enable_xray_mode )
			glDisable( GL_DEPTH_TEST );

		// 位置操作のための座標系を描画
		if ( ( curr_operation == OP_W_TRANSLATION ) || ( curr_operation == OP_M_TRANSLATION ) )
		{
			glPushMatrix();

			// モデル座標系とワールド座標系のどちらで描画するかを設定
			if ( curr_operation == OP_W_TRANSLATION )
				glTranslatef( o.pos );
			if ( curr_operation == OP_M_TRANSLATION )
				glMultMatrixf( frame );

			// 移動を行うXYZ軸を描画
			glLineWidth( ( active_handle == H_W_X_AXIS ) || ( active_handle == H_M_X_AXIS ) ? ro.selected_axis_width : ro.axis_width );
			glBegin( GL_LINES );
				glColor3f( ro.x_axis_color );
				glVertex3f( 0.0f, 0.0f, 0.0f );
				glVertex3f( sx, 0.0f, 0.0f );
			glEnd();

			glLineWidth( ( active_handle == H_W_Y_AXIS ) || ( active_handle == H_M_Y_AXIS ) ? ro.selected_axis_width : ro.axis_width );
			glBegin( GL_LINES );
				glColor3f( ro.y_axis_color );
				glVertex3f( 0.0f, 0.0f, 0.0f );
				glVertex3f( 0.0f, sy, 0.0f );
			glEnd();

			glLineWidth( ( active_handle == H_W_Z_AXIS ) || ( active_handle == H_M_Z_AXIS ) ? ro.selected_axis_width : ro.axis_width );
			glBegin( GL_LINES );
				glColor3f( ro.z_axis_color );
				glVertex3f( 0.0f, 0.0f, 0.0f );
				glVertex3f( 0.0f, 0.0f, sz );
			glEnd();

			glPopMatrix();
		}
		// 回転操作のための座標系を描画
		else if ( ( curr_operation == OP_W_ROTATION ) || ( curr_operation == OP_M_ROTATION ) )
		{
			glPushMatrix();

			// モデル座標系とワールド座標系のどちらで描画するかを設定
			if ( curr_operation == OP_W_ROTATION )
				glTranslatef( o.pos );
			if ( curr_operation == OP_M_ROTATION )
				glMultMatrixf( frame );

			// XYZ軸を描画
			glLineWidth( ro.axis_width );
			glColor3f( 0.0f, 0.0f, 0.0f );
			glBegin( GL_LINES );
				glVertex3f( 0.0f, 0.0f, 0.0f );
				glVertex3f( sx, 0.0f, 0.0f );
				glVertex3f( 0.0f, 0.0f, 0.0f );
				glVertex3f( 0.0f, sy, 0.0f );
				glVertex3f( 0.0f, 0.0f, 0.0f );
				glVertex3f( 0.0f, 0.0f, sz );
			glEnd();

			int  i;
			const int  num_arc_points = 6;
			Point2f    arc_points[ num_arc_points ];
			for ( i=0; i<num_arc_points; i++ )
			{
				float  angle = ( i * 10 + 20 ) * M_PI / 180.0f;
				arc_points[ i ].x = cos( angle );
				arc_points[ i ].y = sin( angle );
			}

			// 回転を行う方向を表す円弧を描画
			glLineWidth( ( active_handle == H_M_YZ_PLANE ) || ( active_handle == H_W_YZ_PLANE ) ? ro.selected_axis_width : ro.axis_width );
			glBegin( GL_LINE_STRIP );
				glColor3f( ro.x_axis_color );
				for ( i=0; i<num_arc_points; i++ )
					glVertex3f( 0.0f, arc_points[ i ].x * sy, arc_points[ i ].y * sz );
			glEnd();

			glLineWidth( ( active_handle == H_M_ZX_PLANE ) || ( active_handle == H_W_ZX_PLANE ) ? ro.selected_axis_width : ro.axis_width );
			glBegin( GL_LINE_STRIP );
				glColor3f( ro.y_axis_color );
				for ( i=0; i<num_arc_points; i++ )
					glVertex3f( arc_points[ i ].x * sx, 0.0f, arc_points[ i ].y * sz );
			glEnd();

			glLineWidth( ( active_handle == H_M_XY_PLANE ) || ( active_handle == H_W_XY_PLANE ) ? ro.selected_axis_width : ro.axis_width );
			glBegin( GL_LINE_STRIP );
				glColor3f( ro.z_axis_color );
				for ( i=0; i<num_arc_points; i++ )
					glVertex3f( arc_points[ i ].x * sx, arc_points[ i ].y * sy, 0.0f );
			glEnd();

			glPopMatrix();
		}
	}

	// レンダリング設定を復元
	if ( b_lighting )
		glEnable( GL_LIGHTING );
	if ( b_depth_test )
		glEnable( GL_DEPTH_TEST );
	else
		glDisable( GL_DEPTH_TEST );
	if ( !b_blend )
		glDisable( GL_BLEND );
	if ( !b_line_smooth )
		glDisable( GL_LINE_SMOOTH );
}



//////////////////////////////////////////////////////////////////////////////
//
//  内部メソッド
//


//
//  全オブジェクトをスクリーン座標系に投影
//
void  ObjectLayout::ProjectObjects()
{
	// OpenGL の変換行列を取得
	double  model_view_matrix[ 16 ];
	double  projection_matrix[ 16 ];
	int  viewport_param[ 4 ];
	Point3d  projected_pos;
	glGetDoublev( GL_MODELVIEW_MATRIX, model_view_matrix );
	glGetDoublev( GL_PROJECTION_MATRIX, projection_matrix );
	glGetIntegerv( GL_VIEWPORT, viewport_param );

	// ワールド座標をスクリーン座標に投影するためのマクロ
	#define  PROJECT( wp, sp )	gluProject( wp.x, wp.y, wp.z, \
			model_view_matrix, projection_matrix, viewport_param, \
			&projected_pos.x, &projected_pos.y, &projected_pos.z ); \
		sp.x = projected_pos.x; \
		sp.y = viewport_param[ 3 ] - projected_pos.y;

	// 各オブジェクトの位置をスクリーン座標に投影
	for ( int i=0; i<objects.size(); i++ )
	{
		Object &  o = objects[ i ];
		PROJECT( o.pos, o.screen_pos );
	}

	// 各操作ハンドルの位置をスクリーン座標に投影
	if ( curr_object != -1 )
	{
		Object &  o = objects[ curr_object ];
		Vector3f  x_axis, y_axis, z_axis, pos;

		// 各軸の長さ（全軸で共通の長さを使用 or 物体のサイズを使用）
		float  sx, sy, sz;
		if ( render_option.change_axis_length )
		{
			sx = o.size.x;  sy = o.size.y;  sz = o.size.z;
		}
		else
		{
			sx = o.axis_length;  sy = o.axis_length;  sz = o.axis_length;
		}

		handle_pos[ H_MODE_CHANGE ] = o.screen_pos;

		o.ori.getColumn( 0, &x_axis );
		o.ori.getColumn( 1, &y_axis );
		o.ori.getColumn( 2, &z_axis );
		pos.scaleAdd( sx, x_axis, o.pos );
		PROJECT( pos, handle_pos[ H_M_X_AXIS ] );
		pos.scaleAdd( sy, y_axis, o.pos );
		PROJECT( pos, handle_pos[ H_M_Y_AXIS ] );
		pos.scaleAdd( sz, z_axis, o.pos );
		PROJECT( pos, handle_pos[ H_M_Z_AXIS ] );
		pos.scaleAdd( sx * 0.7f, x_axis, o.pos );
		pos.scaleAdd( sy * 0.7f, y_axis, pos );
		PROJECT( pos, handle_pos[ H_M_XY_PLANE ] );
		pos.scaleAdd( sy * 0.7f, y_axis, o.pos );
		pos.scaleAdd( sz * 0.7f, z_axis, pos );
		PROJECT( pos, handle_pos[ H_M_YZ_PLANE ] );
		pos.scaleAdd( sz * 0.7f, z_axis, o.pos );
		pos.scaleAdd( sx * 0.7f, x_axis, pos );
		PROJECT( pos, handle_pos[ H_M_ZX_PLANE ] );

		x_axis.set( 1.0f, 0.0f, 0.0f );
		y_axis.set( 0.0f, 1.0f, 0.0f );
		z_axis.set( 0.0f, 0.0f, 1.0f );
		pos.scaleAdd( sx, x_axis, o.pos );
		PROJECT( pos, handle_pos[ H_W_X_AXIS ] );
		pos.scaleAdd( sy, y_axis, o.pos );
		PROJECT( pos, handle_pos[ H_W_Y_AXIS ] );
		pos.scaleAdd( sz, z_axis, o.pos );
		PROJECT( pos, handle_pos[ H_W_Z_AXIS ] );
		pos.scaleAdd( sx * 0.7f, x_axis, o.pos );
		pos.scaleAdd( sy * 0.7f, y_axis, pos );
		PROJECT( pos, handle_pos[ H_W_XY_PLANE ] );
		pos.scaleAdd( sy * 0.7f, y_axis, o.pos );
		pos.scaleAdd( sz * 0.7f, z_axis, pos );
		PROJECT( pos, handle_pos[ H_W_YZ_PLANE ] );
		pos.scaleAdd( sz * 0.7f, z_axis, o.pos );
		pos.scaleAdd( sx * 0.7f, x_axis, pos );
		PROJECT( pos, handle_pos[ H_W_ZX_PLANE ] );
	}
}


//
//  マウス位置に近いオブジェクトを探索
//
int  ObjectLayout::FindObject( int mx, int my, int threshold )
{
	int  closest_no = -1;
	int  closest_dist;
	int  dist, dx, dy;

	// 各オブジェクトのマウス位置の距離を計算して、最も近いものを探索
	for ( int i=0; i<objects.size(); i++ )
	{
		// 距離を計算
		const Object &  o = objects[ i ];
		dx = o.screen_pos.x - mx;
		dy = o.screen_pos.y - my;
		dist = dx * dx + dy * dy;

		// 閾値以下でなければスキップ
		if ( dist > threshold * threshold )
			continue;

		// これまでに見つかったオブジェクトより近ければ更新
		if ( ( closest_no == -1 ) || ( dist < closest_dist ) )
		{
			closest_no = i;
			closest_dist = dist;
		}
	}

	// 見つかったオブジェクト番号を返す
	return  closest_no;
}


//
//  マウス位置に近いハンドルを探索
//
ObjectLayout::HandleEnum  ObjectLayout::FindHandle( int mx, int my, OperationEnum op, int threshold )
{
	int  closest_no = H_NONE;
	int  closest_dist;
	int  dist, dx, dy;

	// 各ハンドルとマウス位置の間の距離を計算して、最も近いハンドルを探索
	for ( int i=0; i<NUM_HANDLE; i++ )
	{
		// 現在の操作が対象とするハンドルでなければスキップ
		if ( ( ( op == OP_W_TRANSLATION ) || ( op == OP_W_ROTATION ) ) &&
		     ( i >= H_M_X_AXIS ) && ( i <= H_M_ZX_PLANE ) )
			continue;
		if ( ( ( op == OP_M_TRANSLATION ) || ( op == OP_M_ROTATION ) ) &&
		     ( i >= H_W_X_AXIS ) && ( i <= H_W_ZX_PLANE ) )
			continue;

		// 距離を計算（暫定）
		dx = handle_pos[ i ].x - mx;
		dy = handle_pos[ i ].y - my;
		dist = dx * dx + dy * dy;

		// 閾値以下でなければスキップ
		if ( dist > threshold * threshold )
			continue;

		// これまでに見つかったオブジェクトより近ければ更新
		if ( ( closest_no == -1 ) || ( dist < closest_dist ) )
		{
			closest_no = i;
			closest_dist = dist;
		}
	}

	// 見つかったハンドル番号を返す
	return  (HandleEnum) closest_no;
}



// End of ObjectLayout.cpp

